@@ -33,13 +33,11 @@ import os
 import json
 import shutil
 import hashlib
-import inspect
 import logging
 import tempfile
 import warnings
 import contextlib
 from pathlib import Path
-from functools import wraps
 from urllib.request import urlopen
 
 import pytest
@@ -83,6 +81,35 @@ def pathify(path):
     return Path(path + ext)
 
 
+def generate_test_name(item):
+    """
+    Generate a unique name for the hash for this test.
+    """
+    if item.cls is not None:
+        name = f"{item.module.__name__}.{item.cls.__name__}.{item.name}"
+    else:
+        name = f"{item.module.__name__}.{item.name}"
+    return name
+
+
+def wrap_figure_interceptor(plugin, item):
+    """
+    Intercept and store figures returned by test functions.
+    """
+    # Only intercept figures on marked figure tests
+    if get_compare(item) is not None:
+
+        # Use the full test name as a key to ensure correct figure is being retrieved
+        test_name = generate_test_name(item)
+
+        def figure_interceptor(store, obj):
+            def wrapper(*args, **kwargs):
+                store.return_value[test_name] = obj(*args, **kwargs)
+            return wrapper
+
+        item.obj = figure_interceptor(plugin, item.obj)
+
+
 def pytest_report_header(config, startdir):
     import matplotlib
     import matplotlib.ft2font
@@ -211,13 +238,11 @@ def close_mpl_figure(fig):
         plt.close(fig)
 
 
-def get_marker(item, marker_name):
-    if hasattr(item, 'get_closest_marker'):
-        return item.get_closest_marker(marker_name)
-    else:
-        # "item.keywords.get" was deprecated in pytest 3.6
-        # See https://docs.pytest.org/en/latest/mark.html#updating-code
-        return item.keywords.get(marker_name)
+def get_compare(item):
+    """
+    Return the mpl_image_compare marker for the given item.
+    """
+    return item.get_closest_marker("mpl_image_compare")
 
 
 def path_is_not_none(apath):
@@ -271,6 +296,7 @@ class ImageComparison:
         self._generated_hash_library = {}
         self._test_results = {}
         self._test_stats = None
+        self.return_value = {}
 
         # https://stackoverflow.com/questions/51737378/how-should-i-log-in-my-pytest-plugin
         # turn debug prints on only if "-vv" or more passed
@@ -278,20 +304,14 @@ class ImageComparison:
         logging.basicConfig(level=level)
         self.logger = logging.getLogger('pytest-mpl')
 
-    def get_compare(self, item):
-        """
-        Return the mpl_image_compare marker for the given item.
-        """
-        return get_marker(item, 'mpl_image_compare')
-
     def generate_filename(self, item):
         """
         Given a pytest item, generate the figure filename.
         """
         if self.config.getini('mpl-use-full-test-name'):
-            filename = self.generate_test_name(item) + '.png'
+            filename = generate_test_name(item) + '.png'
         else:
-            compare = self.get_compare(item)
+            compare = get_compare(item)
             # Find test name to use as plot name
             filename = compare.kwargs.get('filename', None)
             if filename is None:
@@ -300,17 +320,11 @@ class ImageComparison:
         filename = str(pathify(filename))
         return filename
 
-    def generate_test_name(self, item):
-        """
-        Generate a unique name for the hash for this test.
-        """
-        return f"{item.module.__name__}.{item.name}"
-
     def make_test_results_dir(self, item):
         """
         Generate the directory to put the results in.
         """
-        test_name = pathify(self.generate_test_name(item))
+        test_name = pathify(generate_test_name(item))
         results_dir = self.results_dir / test_name
         results_dir.mkdir(exist_ok=True, parents=True)
         return results_dir
@@ -319,7 +333,7 @@ class ImageComparison:
         """
         Returns `True` if a non-default baseline directory is specified.
         """
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         item_baseline_dir = compare.kwargs.get('baseline_dir', None)
         return item_baseline_dir or self.baseline_dir or self.baseline_relative_dir
 
@@ -330,7 +344,7 @@ class ImageComparison:
         Using the global and per-test configuration return the absolute
         baseline dir, if the baseline file is local else return base URL.
         """
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         baseline_dir = compare.kwargs.get('baseline_dir', None)
         if baseline_dir is None:
             if self.baseline_dir is None:
@@ -394,7 +408,7 @@ class ImageComparison:
         """
         Generate reference figures.
         """
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
 
         if not os.path.exists(self.generate_dir):
@@ -413,7 +427,7 @@ class ImageComparison:
         For a `matplotlib.figure.Figure`, returns the SHA256 hash as a hexadecimal
         string.
         """
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
 
         imgdata = io.BytesIO()
@@ -436,7 +450,7 @@ class ImageComparison:
         if summary is None:
             summary = {}
 
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         tolerance = compare.kwargs.get('tolerance', 2)
         savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
 
@@ -510,7 +524,7 @@ class ImageComparison:
         if summary is None:
             summary = {}
 
-        compare = self.get_compare(item)
+        compare = get_compare(item)
         savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
 
         if not self.results_hash_library_name:
@@ -524,7 +538,7 @@ class ImageComparison:
             pytest.fail(f"Can't find hash library at path {hash_library_filename}")
 
         hash_library = self.load_hash_library(hash_library_filename)
-        hash_name = self.generate_test_name(item)
+        hash_name = generate_test_name(item)
         baseline_hash = hash_library.get(hash_name, None)
         summary['baseline_hash'] = baseline_hash
 
@@ -582,11 +596,13 @@ class ImageComparison:
             return
         return summary['status_msg']
 
-    def pytest_runtest_setup(self, item):  # noqa
+    @pytest.hookimpl(hookwrapper=True)
+    def pytest_runtest_call(self, item):  # noqa
 
-        compare = self.get_compare(item)
+        compare = get_compare(item)
 
         if compare is None:
+            yield
             return
 
         import matplotlib.pyplot as plt
@@ -600,95 +616,82 @@ class ImageComparison:
         remove_text = compare.kwargs.get('remove_text', False)
         backend = compare.kwargs.get('backend', 'agg')
 
-        original = item.function
+        with plt.style.context(style, after_reset=True), switch_backend(backend):
 
-        @wraps(item.function)
-        def item_function_wrapper(*args, **kwargs):
+            # Run test and get figure object
+            wrap_figure_interceptor(self, item)
+            yield
+            test_name = generate_test_name(item)
+            if test_name not in self.return_value:
+                # Test function did not complete successfully
+                return
+            fig = self.return_value[test_name]
 
-            with plt.style.context(style, after_reset=True), switch_backend(backend):
+            if remove_text:
+                remove_ticks_and_titles(fig)
 
-                # Run test and get figure object
-                if inspect.ismethod(original):  # method
-                    # In some cases, for example if setup_method is used,
-                    # original appears to belong to an instance of the test
-                    # class that is not the same as args[0], and args[0] is the
-                    # one that has the correct attributes set up from setup_method
-                    # so we ignore original.__self__ and use args[0] instead.
-                    fig = original.__func__(*args, **kwargs)
-                else:  # function
-                    fig = original(*args, **kwargs)
+            result_dir = self.make_test_results_dir(item)
 
-                if remove_text:
-                    remove_ticks_and_titles(fig)
+            summary = {
+                'status': None,
+                'image_status': None,
+                'hash_status': None,
+                'status_msg': None,
+                'baseline_image': None,
+                'diff_image': None,
+                'rms': None,
+                'tolerance': None,
+                'result_image': None,
+                'baseline_hash': None,
+                'result_hash': None,
+            }
 
-                test_name = self.generate_test_name(item)
-                result_dir = self.make_test_results_dir(item)
+            # What we do now depends on whether we are generating the
+            # reference images or simply running the test.
+            if self.generate_dir is not None:
+                summary['status'] = 'skipped'
+                summary['image_status'] = 'generated'
+                summary['status_msg'] = 'Skipped test, since generating image.'
+                generate_image = self.generate_baseline_image(item, fig)
+                if self.results_always:  # Make baseline image available in HTML
+                    result_image = (result_dir / "baseline.png").absolute()
+                    shutil.copy(generate_image, result_image)
+                    summary['baseline_image'] = \
+                        result_image.relative_to(self.results_dir).as_posix()
 
-                summary = {
-                    'status': None,
-                    'image_status': None,
-                    'hash_status': None,
-                    'status_msg': None,
-                    'baseline_image': None,
-                    'diff_image': None,
-                    'rms': None,
-                    'tolerance': None,
-                    'result_image': None,
-                    'baseline_hash': None,
-                    'result_hash': None,
-                }
+            if self.generate_hash_library is not None:
+                summary['hash_status'] = 'generated'
+                image_hash = self.generate_image_hash(item, fig)
+                self._generated_hash_library[test_name] = image_hash
+                summary['baseline_hash'] = image_hash
 
-                # What we do now depends on whether we are generating the
-                # reference images or simply running the test.
-                if self.generate_dir is not None:
-                    summary['status'] = 'skipped'
-                    summary['image_status'] = 'generated'
-                    summary['status_msg'] = 'Skipped test, since generating image.'
-                    generate_image = self.generate_baseline_image(item, fig)
-                    if self.results_always:  # Make baseline image available in HTML
-                        result_image = (result_dir / "baseline.png").absolute()
-                        shutil.copy(generate_image, result_image)
-                        summary['baseline_image'] = \
-                            result_image.relative_to(self.results_dir).as_posix()
+            # Only test figures if not generating images
+            if self.generate_dir is None:
+                # Compare to hash library
+                if self.hash_library or compare.kwargs.get('hash_library', None):
+                    msg = self.compare_image_to_hash_library(item, fig, result_dir, summary=summary)
 
-                if self.generate_hash_library is not None:
-                    summary['hash_status'] = 'generated'
-                    image_hash = self.generate_image_hash(item, fig)
-                    self._generated_hash_library[test_name] = image_hash
-                    summary['baseline_hash'] = image_hash
-
-                # Only test figures if not generating images
-                if self.generate_dir is None:
-                    # Compare to hash library
-                    if self.hash_library or compare.kwargs.get('hash_library', None):
-                        msg = self.compare_image_to_hash_library(item, fig, result_dir, summary=summary)
-
-                    # Compare against a baseline if specified
-                    else:
-                        msg = self.compare_image_to_baseline(item, fig, result_dir, summary=summary)
-
-                    close_mpl_figure(fig)
-
-                    if msg is None:
-                        if not self.results_always:
-                            shutil.rmtree(result_dir)
-                            for image_type in ['baseline_image', 'diff_image', 'result_image']:
-                                summary[image_type] = None  # image no longer exists
-                    else:
-                        self._test_results[test_name] = summary
-                        pytest.fail(msg, pytrace=False)
+                # Compare against a baseline if specified
+                else:
+                    msg = self.compare_image_to_baseline(item, fig, result_dir, summary=summary)
 
                 close_mpl_figure(fig)
 
-                self._test_results[test_name] = summary
+                if msg is None:
+                    if not self.results_always:
+                        shutil.rmtree(result_dir)
+                        for image_type in ['baseline_image', 'diff_image', 'result_image']:
+                            summary[image_type] = None  # image no longer exists
+                else:
+                    self._test_results[test_name] = summary
+                    pytest.fail(msg, pytrace=False)
 
-                if summary['status'] == 'skipped':
-                    pytest.skip(summary['status_msg'])
+            close_mpl_figure(fig)
 
-        if item.cls is not None:
-            setattr(item.cls, item.function.__name__, item_function_wrapper)
-        else:
-            item.obj = item_function_wrapper
+            self._test_results[test_name] = summary
+
+            if summary['status'] == 'skipped':
+                pytest.skip(summary['status_msg'])
 
     def generate_summary_json(self):
         json_file = self.results_dir / 'results.json'
@@ -726,11 +729,11 @@ class ImageComparison:
                 kwargs["hash_library"] = result_hash_library.name
             if 'html' in self.generate_summary:
                 summary = generate_summary_html(self._test_results, self.results_dir, **kwargs)
-                print(f"A summary of the failed tests can be found at: {summary}")
+                print(f"A summary of test results can be found at: {summary}")
             if 'basic-html' in self.generate_summary:
                 summary = generate_summary_basic_html(self._test_results, self.results_dir,
                                                       **kwargs)
-                print(f"A summary of the failed tests can be found at: {summary}")
+                print(f"A summary of test results can be found at: {summary}")
 
 
 class FigureCloser:
@@ -741,27 +744,16 @@ class FigureCloser:
 
     def __init__(self, config):
         self.config = config
+        self.return_value = {}
 
-    def pytest_runtest_setup(self, item):
-
-        compare = get_marker(item, 'mpl_image_compare')
-
-        if compare is None:
-            return
-
-        original = item.function
-
-        @wraps(item.function)
-        def item_function_wrapper(*args, **kwargs):
-
-            if inspect.ismethod(original):  # method
-                fig = original.__func__(*args, **kwargs)
-            else:  # function
-                fig = original(*args, **kwargs)
-
+    @pytest.hookimpl(hookwrapper=True)
+    def pytest_runtest_call(self, item):
+        wrap_figure_interceptor(self, item)
+        yield
+        if get_compare(item) is not None:
+            test_name = generate_test_name(item)
+            if test_name not in self.return_value:
+                # Test function did not complete successfully
+                return
+            fig = self.return_value[test_name]
             close_mpl_figure(fig)
-
-        if item.cls is not None:
-            setattr(item.cls, item.function.__name__, item_function_wrapper)
-        else:
-            item.obj = item_function_wrapper
